# Copyright Teklia (contact@teklia.com) & Denis Coquenet
# This code is licensed under CeCILL-C

# -*- coding: utf-8 -*-

from pathlib import Path

FIXTURES = Path(__file__).resolve().parent / "data"



# Copyright Teklia (contact@teklia.com) & Denis Coquenet
# This code is licensed under CeCILL-C

import logging

import pytest

from dan.bio import convert
from dan.utils import EntityType

ST_TEXT = """ⒶBryan B ⒷParis ⒸJanuary 1st, 1987
ⒶJoe J ⒷGrenoble ⒸAugust 24, 1995
ⒶHannah H ⒷLille ⒸSeptember 15, 2002"""

ST_ET_TEXT = """ⒶBryanⒷ and ⒶJoeⒷ will visit the ⒸEiffel TowerⒹ in ⒸParisⒹ next ⒺTuesdayⒻ.
ⒶHannahⒷ will visit the ⒸPlace ⒶCharles de GaulleⒷ étoileⒹ on ⒺWednesdayⒻ."""


def test_convert_with_error():
    ner_tokens = {
        "Person": EntityType(start="Ⓐ", end="Ⓑ"),
        "Location": EntityType(start="Ⓒ", end="Ⓓ"),
    }

    with pytest.raises(
        AssertionError, match="Ending token Ⓓ doesn't match the starting token Ⓐ"
    ):
        convert("ⒶFredⒹ", ner_tokens)


def test_convert_with_warnings(caplog):
    ner_tokens = {
        "Person": EntityType(start="Ⓐ", end="Ⓑ"),
        "Location": EntityType(start="Ⓒ", end="Ⓓ"),
    }

    assert convert("BryanⒷ and ⒶJoeⒷ will visit the Eiffel TowerⒹ", ner_tokens).split(
        "\n"
    ) == [
        "Bryan O",
        "and O",
        "Joe B-Person",
        "will O",
        "visit O",
        "the O",
        "Eiffel O",
        "Tower O",
    ]
    assert [(level, message) for _, level, message in caplog.record_tuples] == [
        (
            logging.WARNING,
            "Missing starting token for ending token Ⓑ, skipping the entity",
        ),
        (
            logging.WARNING,
            "Missing starting token for ending token Ⓓ, skipping the entity",
        ),
    ]


def test_convert_starting_tokens():
    ner_tokens = {
        "Person": EntityType(start="Ⓐ"),
        "Location": EntityType(start="Ⓑ"),
        "Date": EntityType(start="Ⓒ"),
    }

    assert convert(ST_TEXT, ner_tokens).split("\n") == [
        "Bryan B-Person",
        "B I-Person",
        "Paris B-Location",
        "January B-Date",
        "1st, I-Date",
        "1987 I-Date",
        "Joe B-Person",
        "J I-Person",
        "Grenoble B-Location",
        "August B-Date",
        "24, I-Date",
        "1995 I-Date",
        "Hannah B-Person",
        "H I-Person",
        "Lille B-Location",
        "September B-Date",
        "15, I-Date",
        "2002 I-Date",
    ]


def test_convert_starting_and_ending_tokens():
    ner_tokens = {
        "Person": EntityType(start="Ⓐ", end="Ⓑ"),
        "Location": EntityType(start="Ⓒ", end="Ⓓ"),
        "Date": EntityType(start="Ⓔ", end="Ⓕ"),
    }

    assert convert(ST_ET_TEXT, ner_tokens).split("\n") == [
        "Bryan B-Person",
        "and O",
        "Joe B-Person",
        "will O",
        "visit O",
        "the O",
        "Eiffel B-Location",
        "Tower I-Location",
        "in O",
        "Paris B-Location",
        "next O",
        "Tuesday B-Date",
        ". O",
        "Hannah B-Person",
        "will O",
        "visit O",
        "the O",
        "Place B-Location",
        "Charles B-Person",
        "de I-Person",
        "Gaulle I-Person",
        "étoile I-Location",
        "on O",
        "Wednesday B-Date",
        ". O",
    ]



# Copyright Teklia (contact@teklia.com) & Denis Coquenet
# This code is licensed under CeCILL-C

import pytest
import yaml

from dan.utils import (
    EntityType,
    parse_charset_pattern,
    parse_tokens,
    parse_tokens_pattern,
)


@pytest.mark.parametrize(
    ("tokens", "error_msg"),
    [
        # All should have no end tokens
        (
            {
                "name": {
                    "start": "A",
                    "end": "",
                },
                "surname": {"start": "B", "end": "C"},
            },
            "Some entities have end tokens",
        ),
        # All should have end tokens
        (
            {
                "name": {
                    "start": "A",
                    "end": "C",
                },
                "surname": {"start": "B", "end": ""},
            },
            "Some entities have no end token",
        ),
    ],
)
def test_parse_tokens_errors(tmp_path, tokens, error_msg):
    tokens_path = tmp_path / "tokens.yml"
    tokens_path.write_text(yaml.dump(tokens))

    with pytest.raises(AssertionError, match=error_msg):
        parse_tokens(tokens_path)


@pytest.mark.parametrize(
    ("pattern", "entity_types"),
    [
        # No end tokens
        (
            r"([ⒶⒷⒸ][^ⒶⒷⒸ]*)",
            [EntityType(start="Ⓐ"), EntityType(start="Ⓑ"), EntityType(start="Ⓒ")],
        ),
        # With end tokens
        (
            r"(ⓐ.*?Ⓐ)|(ⓑ.*?Ⓑ)|(ⓒ.*?Ⓒ)",
            [
                EntityType(start="ⓐ", end="Ⓐ"),
                EntityType(start="ⓑ", end="Ⓑ"),
                EntityType(start="ⓒ", end="Ⓒ"),
            ],
        ),
    ],
)
def test_parse_tokens_pattern(pattern: str, entity_types: list[EntityType]):
    assert parse_tokens_pattern(entity_types).pattern == pattern


@pytest.mark.parametrize(
    ("charset", "pattern"),
    [
        (["a", "b", "c"], r"[abc]"),
        (["^", "a", "b", "c"], r"[\^abc]"),
        (
            ["<language>", "[", "<word>", "]", "\\", '"'],
            r'(?:<language>)|(?:<word>)|[\[\]\\"]',
        ),
    ],
)
def test_parse_charset_pattern(charset, pattern):
    assert parse_charset_pattern(charset).pattern == pattern



# Copyright Teklia (contact@teklia.com) & Denis Coquenet
# This code is licensed under CeCILL-C

# -*- coding: utf-8 -*-

import json
import logging
from operator import attrgetter, methodcaller
from pathlib import Path

import pytest
from PIL import Image, ImageChops

from dan.datasets.download.images import IIIF_FULL_SIZE, ImageDownloader
from dan.datasets.download.utils import download_image
from line_image_extractor.image_utils import BoundingBox
from tests import FIXTURES

EXTRACTION_DATA_PATH = FIXTURES / "extraction"


@pytest.mark.parametrize(
    "max_width, max_height, width, height, resize",
    (
        (1000, 2000, 900, 800, IIIF_FULL_SIZE),
        (1000, 2000, 1100, 800, "1000,"),
        (1000, 2000, 1100, 2800, ",2000"),
        (1000, 2000, 2000, 3000, "1000,"),
    ),
)
def test_get_iiif_size_arg(max_width, max_height, width, height, resize):
    assert (
        ImageDownloader(max_width=max_width, max_height=max_height).get_iiif_size_arg(
            width=width, height=height
        )
        == resize
    )


def test_download(split_content, monkeypatch, tmp_path):
    # Mock download_image so that it simply opens it with Pillow
    monkeypatch.setattr(
        "dan.datasets.download.images.download_image", lambda url: Image.open(url)
    )

    output = tmp_path / "download"
    output.mkdir(parents=True, exist_ok=True)
    (output / "split.json").write_text(json.dumps(split_content))

    def mock_build_image_url(polygon, image_url, *args, **kwargs):
        # During tests, the image URL is its local path
        return image_url

    extractor = ImageDownloader(
        output=output,
        image_extension=".jpg",
    )
    # Mock build_image_url to simply return the path to the image
    extractor.build_iiif_url = mock_build_image_url
    extractor.run()

    # Check files
    IMAGE_DIR = output / "images"
    TEST_DIR = IMAGE_DIR / "test" / "dataset_id"
    TRAIN_DIR = IMAGE_DIR / "train" / "dataset_id"
    VAL_DIR = IMAGE_DIR / "val" / "dataset_id"

    expected_paths = [
        # Images of test folder
        TEST_DIR / "test-page_1-line_1.jpg",
        TEST_DIR / "test-page_1-line_2.jpg",
        TEST_DIR / "test-page_1-line_3.jpg",
        TEST_DIR / "test-page_2-line_1.jpg",
        TEST_DIR / "test-page_2-line_2.jpg",
        TEST_DIR / "test-page_2-line_3.jpg",
        # Images of train folder
        TRAIN_DIR / "train-page_1-line_1.jpg",
        TRAIN_DIR / "train-page_1-line_2.jpg",
        TRAIN_DIR / "train-page_1-line_3.jpg",
        TRAIN_DIR / "train-page_1-line_4.jpg",
        TRAIN_DIR / "train-page_2-line_1.jpg",
        TRAIN_DIR / "train-page_2-line_2.jpg",
        TRAIN_DIR / "train-page_2-line_3.jpg",
        # Images of val folder
        VAL_DIR / "val-page_1-line_1.jpg",
        VAL_DIR / "val-page_1-line_2.jpg",
        VAL_DIR / "val-page_1-line_3.jpg",
        output / "labels.json",
        output / "split.json",
    ]
    assert sorted(filter(methodcaller("is_file"), output.rglob("*"))) == expected_paths

    # Check "labels.json"
    expected_labels = {
        "test": {
            "images/test/dataset_id/test-page_1-line_1.jpg": "ⓢCou⁇e⁇  ⓕBouis  ⓑ⁇.12.14",
            "images/test/dataset_id/test-page_1-line_2.jpg": "ⓢ⁇outrain  ⓕA⁇ol⁇⁇e  ⓑ9.4.13",
            "images/test/dataset_id/test-page_1-line_3.jpg": "ⓢ⁇abale  ⓕ⁇ran⁇ais  ⓑ26.3.11",
            "images/test/dataset_id/test-page_2-line_1.jpg": "ⓢ⁇urosoy  ⓕBouis  ⓑ22⁇4⁇18",
            "images/test/dataset_id/test-page_2-line_2.jpg": "ⓢColaiani  ⓕAn⁇els  ⓑ28.11.1⁇",
            "images/test/dataset_id/test-page_2-line_3.jpg": "ⓢRenouar⁇  ⓕMaurice  ⓑ2⁇.⁇.04",
        },
        "train": {
            "images/train/dataset_id/train-page_1-line_1.jpg": "ⓢCaillet  ⓕMaurice  ⓑ28.9.06",
            "images/train/dataset_id/train-page_1-line_2.jpg": "ⓢReboul  ⓕJean  ⓑ30.9.02",
            "images/train/dataset_id/train-page_1-line_3.jpg": "ⓢBareyre  ⓕJean  ⓑ28.3.11",
            "images/train/dataset_id/train-page_1-line_4.jpg": "ⓢRoussy  ⓕJean  ⓑ4.11.14",
            "images/train/dataset_id/train-page_2-line_1.jpg": "ⓢMarin  ⓕMarcel  ⓑ10.8.06",
            "images/train/dataset_id/train-page_2-line_2.jpg": "ⓢAmical  ⓕEloi  ⓑ11.10.04",
            "images/train/dataset_id/train-page_2-line_3.jpg": "ⓢBiros  ⓕMael  ⓑ30.10.10",
        },
        "val": {
            "images/val/dataset_id/val-page_1-line_1.jpg": "ⓢMonar⁇  ⓕBouis  ⓑ29⁇⁇⁇04",
            "images/val/dataset_id/val-page_1-line_2.jpg": "ⓢAstier  ⓕArt⁇ur  ⓑ11⁇2⁇13",
            "images/val/dataset_id/val-page_1-line_3.jpg": "ⓢ⁇e ⁇lie⁇er  ⓕJules  ⓑ21⁇11⁇11",
        },
    }

    assert json.loads((output / "labels.json").read_text()) == expected_labels

    # Check cropped images
    for expected_path in expected_paths:
        if expected_path.suffix != ".jpg":
            continue

        assert ImageChops.difference(
            Image.open(
                EXTRACTION_DATA_PATH / "images" / "text_line" / expected_path.name
            ),
            Image.open(expected_path),
        )


def test_download_image_error(monkeypatch, caplog, capsys):
    task = {
        "split": "train",
        "polygon": [],
        "image_url": "deadbeef",
        "destination": Path("/dev/null"),
    }
    monkeypatch.setattr(
        "dan.datasets.download.images.polygon_to_bbox",
        lambda polygon: BoundingBox(0, 0, 0, 0),
    )

    extractor = ImageDownloader(image_extension=".jpg")

    # Add the key in data
    extractor.data[task["split"]][str(task["destination"])] = "deadbeefdata"

    # Build a random task
    extractor.download_images([task])

    # Key should have been removed
    assert str(task["destination"]) not in extractor.data[task["split"]]

    # Check error log
    assert len(caplog.record_tuples) == 1
    _, level, msg = caplog.record_tuples[0]
    assert level == logging.ERROR
    assert msg == "Failed to download 1 image(s)."

    # Check stdout
    captured = capsys.readouterr()
    assert captured.out == "deadbeef: Image URL must be HTTP(S) for element null\n"


def test_download_image_error_try_max(responses, caplog):
    # An image's URL
    url = (
        "https://blabla.com/iiif/2/image_path.jpg/231,699,2789,3659/full/0/default.jpg"
    )
    fixed_url = (
        "https://blabla.com/iiif/2/image_path.jpg/231,699,2789,3659/max/0/default.jpg"
    )

    # Fake responses error
    responses.add(
        responses.GET,
        url,
        status=400,
    )
    # Correct response with max
    responses.add(
        responses.GET,
        fixed_url,
        status=200,
        body=next((FIXTURES / "prediction" / "images").iterdir()).read_bytes(),
    )

    image = download_image(url)

    assert image
    # We try 3 times with the first URL
    # Then the first try with the new URL is successful
    assert len(responses.calls) == 4
    assert list(map(attrgetter("request.url"), responses.calls)) == [url] * 3 + [
        fixed_url
    ]

    # Check error log
    assert len(caplog.record_tuples) == 2

    # We should only have WARNING levels
    assert set(level for _, level, _ in caplog.record_tuples) == {logging.WARNING}



# Copyright Teklia (contact@teklia.com) & Denis Coquenet
# This code is licensed under CeCILL-C

# -*- coding: utf-8 -*-

import json
import pickle
import re
from operator import methodcaller
from typing import NamedTuple

import pytest

from arkindex_export import (
    DatasetElement,
    Element,
    Transcription,
    TranscriptionEntity,
)
from dan.datasets.extract.arkindex import ArkindexExtractor
from dan.datasets.extract.db import get_transcription_entities
from dan.datasets.extract.exceptions import (
    NoTranscriptionError,
    UnknownTokenInText,
)
from dan.datasets.extract.utils import (
    EntityType,
    entities_to_xml,
    normalize_linebreaks,
    normalize_spaces,
)
from dan.utils import parse_tokens
from tests import FIXTURES

EXTRACTION_DATA_PATH = FIXTURES / "extraction"

TWO_SPACES_REGEX = re.compile(r" {2}")
ENTITY_TOKEN_SPACE = re.compile(r"[ⓢ|ⓕ|ⓑ] ")
TWO_SPACES_LM_REGEX = re.compile(r"▁ ▁")

# NamedTuple to mock actual database result
Entity = NamedTuple("Entity", offset=int, length=int, type=str, value=str)

TOKENS = {
    "P": EntityType(start="ⓟ", end="Ⓟ"),
    "D": EntityType(start="ⓓ", end="Ⓓ"),
    "N": EntityType(start="ⓝ", end="Ⓝ"),
    "I": EntityType(start="ⓘ", end="Ⓘ"),
}


def filter_tokens(keys):
    return {key: value for key, value in TOKENS.items() if key in keys}


@pytest.mark.parametrize(
    "text,trimmed",
    (
        ("no_spaces", "no_spaces"),
        (" beginning", "beginning"),
        ("ending ", "ending"),
        (" both ", "both"),
        ("    consecutive", "consecutive"),
        ("\ttab", "tab"),
        ("\t tab", "tab"),
        (" \ttab", "tab"),
        ("no|space", "no|space"),
    ),
)
def test_normalize_spaces(text, trimmed):
    assert normalize_spaces(text) == trimmed


@pytest.mark.parametrize(
    "text,trimmed",
    (
        ("no_linebreaks", "no_linebreaks"),
        ("\nbeginning", "beginning"),
        ("ending\n", "ending"),
        ("\nboth\n", "both"),
        ("\n\n\nconsecutive", "consecutive"),
        ("\rcarriage_return", "carriage_return"),
        ("\r\ncarriage_return+linebreak", "carriage_return+linebreak"),
        ("\n\r\r\n\ncarriage_return+linebreak", "carriage_return+linebreak"),
        ("no|linebreaks", "no|linebreaks"),
    ),
)
def test_normalize_linebreaks(text, trimmed):
    assert normalize_linebreaks(text) == trimmed


def test_process_element_unknown_token_in_text_error(mock_database, tmp_path):
    output = tmp_path / "extraction"
    arkindex_extractor = ArkindexExtractor(output=output)

    # Retrieve a dataset element and update its transcription with an invalid one
    dataset_element = DatasetElement.select().first()
    element = dataset_element.element
    Transcription.update({Transcription.text: "Is this text valid⁇"}).execute()

    with pytest.raises(
        UnknownTokenInText,
        match=re.escape(
            f"Unknown token found in the transcription text of element ({element.id})"
        ),
    ):
        arkindex_extractor.process_element(dataset_element, element)


@pytest.mark.parametrize(
    "load_entities,keep_spaces,transcription_entities_worker_version,expected_subword_language_corpus,subword_vocab_size",
    (
        (
            True,
            True,
            "worker_version_id",
            """▁ ⓢ c a i l l e t ▁ ⓕ m a u r i c e ▁ ⓑ 28. 9.0 6
▁ ⓢ re b ou l ▁ ⓕ j e a n ▁ ⓑ 30. 9.0 2
▁ ⓢ b a re y re ▁ ⓕ j e a n ▁ ⓑ 28. 3 . 1 1
▁ ⓢ r ou s s y ▁ ⓕ j e a n ▁ ⓑ 4 . 1 1 . 1 4
▁ ⓢ m a r i n ▁ ⓕ m a r c e l ▁ ⓑ 1 0 . 8 . 0 6
▁ ⓢ a m i c a l ▁ ⓕ e l o i ▁ ⓑ 1 1 . 1 0 . 0 4
▁ ⓢ b i r o s ▁ ⓕ m a e l ▁ ⓑ 30. 1 0 . 1 0""",
            40,
        ),
        (
            True,
            False,
            "worker_version_id",
            """▁ ⓢ c a i l l e t ▁ ⓕ m a u r i c e ▁ ⓑ 28. 9.0 6
▁ ⓢ re b ou l ▁ ⓕ j e a n ▁ ⓑ 30. 9.0 2
▁ ⓢ b a re y re ▁ ⓕ j e a n ▁ ⓑ 28. 3 . 1 1
▁ ⓢ r ou s s y ▁ ⓕ j e a n ▁ ⓑ 4 . 1 1 . 1 4
▁ ⓢ m a r i n ▁ ⓕ m a r c e l ▁ ⓑ 1 0 . 8 . 0 6
▁ ⓢ a m i c a l ▁ ⓕ e l o i ▁ ⓑ 1 1 . 1 0 . 0 4
▁ ⓢ b i r o s ▁ ⓕ m a e l ▁ ⓑ 30. 1 0 . 1 0""",
            40,
        ),
        (
            False,
            True,
            "worker_version_id",
            """▁ ca i l l e t ▁ ma u r i ce ▁ 28. 9.0 6
▁ re b o u l ▁ j e a n ▁ 30. 9.0 2
▁ b a re y re ▁ j e a n ▁ 28. 3 . 1 1
▁ r o u s s y ▁ j e a n ▁ 4 . 11.1 4
▁ ma r i n ▁ ma r ce l ▁ 10. 8 . 0 6
▁ a m i ca l ▁ el o i ▁ 11.1 0 . 0 4
▁ b i r o s ▁ ma el ▁ 30. 10. 1 0""",
            40,
        ),
        (
            False,
            False,
            "worker_version_id",
            """▁ ca i l l e t ▁ ma u r i ce ▁ 28. 9.0 6
▁ re b o u l ▁ j e a n ▁ 30. 9.0 2
▁ b a re y re ▁ j e a n ▁ 28. 3 . 1 1
▁ r o u s s y ▁ j e a n ▁ 4 . 11.1 4
▁ ma r i n ▁ ma r ce l ▁ 10. 8 . 0 6
▁ a m i ca l ▁ el o i ▁ 11.1 0 . 0 4
▁ b i r o s ▁ ma el ▁ 30. 10. 1 0""",
            40,
        ),
        (
            True,
            True,
            False,
            """▁ ⓢ C a i l l e t ▁ ⓕ M a u r i c e ▁ ⓑ 2 8 . 9 . 0 6
▁ ⓢ R e b o u l ▁ ⓕ J e a n ▁ ⓑ 3 0 . 9 . 0 2
▁ ⓢ B a r e y r e ▁ ⓕ J e a n ▁ ⓑ 2 8 . 3 . 1 1
▁ ⓢ R o u s s y ▁ ⓕ J e a n ▁ ⓑ 4 . 1 1 . 1 4
▁ ⓢ M a r i n ▁ ⓕ M a r c e l ▁ ⓑ 1 0 . 8 . 0 6
▁ ⓢ A m i c a l ▁ ⓕ E l o i ▁ ⓑ 1 1 . 1 0 . 0 4
▁ ⓢ B i r o s ▁ ⓕ M a e l ▁ ⓑ 3 0 . 1 0 . 1 0""",
            40,
        ),
        (
            True,
            True,
            False,
            """▁ ⓢ C a i l l e t ▁ ⓕ M a u ri ce ▁ ⓑ 28. 9.0 6
▁ ⓢ R e b ou l ▁ ⓕ J e a n ▁ ⓑ 30. 9.0 2
▁ ⓢ B a re y re ▁ ⓕ J e a n ▁ ⓑ 28. 3 . 1 1
▁ ⓢ R ou s s y ▁ ⓕ J e a n ▁ ⓑ 4 . 11.1 4
▁ ⓢ Mar i n ▁ ⓕ Mar ce l ▁ ⓑ 10. 8 . 0 6
▁ ⓢ A m ic a l ▁ ⓕ E l o i ▁ ⓑ 11.1 0 . 0 4
▁ ⓢ B i r o s ▁ ⓕ M a e l ▁ ⓑ 30. 10. 10""",
            55,
        ),
        (
            True,
            False,
            False,
            """▁ ⓢ C a i l l e t ▁ ⓕ M a u r i c e ▁ ⓑ 2 8 . 9 . 0 6
▁ ⓢ R e b o u l ▁ ⓕ J e a n ▁ ⓑ 3 0 . 9 . 0 2
▁ ⓢ B a r e y r e ▁ ⓕ J e a n ▁ ⓑ 2 8 . 3 . 1 1
▁ ⓢ R o u s s y ▁ ⓕ J e a n ▁ ⓑ 4 . 1 1 . 1 4
▁ ⓢ M a r i n ▁ ⓕ M a r c e l ▁ ⓑ 1 0 . 8 . 0 6
▁ ⓢ A m i c a l ▁ ⓕ E l o i ▁ ⓑ 1 1 . 1 0 . 0 4
▁ ⓢ B i r o s ▁ ⓕ M a e l ▁ ⓑ 3 0 . 1 0 . 1 0""",
            40,
        ),
        (
            False,
            True,
            False,
            """▁ C a i l l e t ▁ Ma u r i c e ▁ 28. 9.0 6
▁ R e b o u l ▁ J e a n ▁ 30. 9.0 2
▁ B a r e y r e ▁ J e a n ▁ 28. 3 . 1 1
▁ R o u s s y ▁ J e a n ▁ 4 . 1 1 . 1 4
▁ Ma r i n ▁ Ma r c e l ▁ 1 0 . 8 . 0 6
▁ A m i c a l ▁ E l o i ▁ 1 1 . 1 0 . 0 4
▁ B i r o s ▁ Ma e l ▁ 30. 1 0 . 1 0""",
            40,
        ),
        (
            False,
            False,
            False,
            """▁ C a i l l e t ▁ Ma u r i c e ▁ 28. 9.0 6
▁ R e b o u l ▁ J e a n ▁ 30. 9.0 2
▁ B a r e y r e ▁ J e a n ▁ 28. 3 . 1 1
▁ R o u s s y ▁ J e a n ▁ 4 . 1 1 . 1 4
▁ Ma r i n ▁ Ma r c e l ▁ 1 0 . 8 . 0 6
▁ A m i c a l ▁ E l o i ▁ 1 1 . 1 0 . 0 4
▁ B i r o s ▁ Ma e l ▁ 30. 1 0 . 1 0""",
            40,
        ),
    ),
)
def test_extract(
    load_entities,
    keep_spaces,
    transcription_entities_worker_version,
    split_content,
    mock_database,
    expected_subword_language_corpus,
    subword_vocab_size,
    tmp_path,
):
    output = tmp_path / "extraction"
    output.mkdir(parents=True, exist_ok=True)
    (output / "language_model").mkdir(parents=True, exist_ok=True)
    tokens_path = EXTRACTION_DATA_PATH / "tokens.yml"
    tokens = [
        token
        for entity_type in parse_tokens(tokens_path).values()
        for token in [entity_type.start, entity_type.end]
        if token
    ]

    extractor = ArkindexExtractor(
        dataset_ids=["dataset_id"],
        element_type=["text_line"],
        output=output,
        # Keep the whole text
        entity_separators=None,
        tokens=tokens_path if load_entities else None,
        transcription_worker_versions=[transcription_entities_worker_version],
        entity_worker_versions=[transcription_entities_worker_version]
        if load_entities
        else [],
        keep_spaces=keep_spaces,
        subword_vocab_size=subword_vocab_size,
    )
    extractor.run()

    expected_paths = [
        output / "charset.pkl",
        # Language resources
        output / "language_model" / "corpus_characters.txt",
        output / "language_model" / "corpus_subwords.txt",
        output / "language_model" / "corpus_words.txt",
        output / "language_model" / "lexicon_characters.txt",
        output / "language_model" / "lexicon_subwords.txt",
        output / "language_model" / "lexicon_words.txt",
        output / "language_model" / "subword_tokenizer.model",
        output / "language_model" / "subword_tokenizer.vocab",
        output / "language_model" / "tokens.txt",
        output / "split.json",
    ]
    assert sorted(filter(methodcaller("is_file"), output.rglob("*"))) == expected_paths

    # Check "split.json"
    # Transcriptions with worker version are in lowercase
    if transcription_entities_worker_version:
        for split in split_content:
            for element_id in split_content[split]:
                split_content[split][element_id]["text"] = split_content[split][
                    element_id
                ]["text"].lower()

    # If we do not load entities, remove tokens
    if not load_entities:
        token_translations = {ord(token): None for token in tokens}
        for split in split_content:
            for element_id in split_content[split]:
                split_content[split][element_id]["text"] = split_content[split][
                    element_id
                ]["text"].translate(token_translations)

    # Replace double spaces with regular space
    if not keep_spaces:
        for split in split_content:
            for element_id in split_content[split]:
                split_content[split][element_id]["text"] = TWO_SPACES_REGEX.sub(
                    " ", split_content[split][element_id]["text"]
                )

    assert json.loads((output / "split.json").read_text()) == split_content

    # Check "charset.pkl"
    expected_charset = set()
    for values in split_content["train"].values():
        expected_charset.update(set(values["text"]))

    if load_entities:
        expected_charset.update(tokens)
    expected_charset.add("⁇")
    assert set(pickle.loads((output / "charset.pkl").read_bytes())) == expected_charset

    # Check "language_corpus.txt"
    expected_char_language_corpus = """ⓢ C a i l l e t ▁ ▁ ⓕ M a u r i c e ▁ ▁ ⓑ 2 8 . 9 . 0 6
ⓢ R e b o u l ▁ ▁ ⓕ J e a n ▁ ▁ ⓑ 3 0 . 9 . 0 2
ⓢ B a r e y r e ▁ ▁ ⓕ J e a n ▁ ▁ ⓑ 2 8 . 3 . 1 1
ⓢ R o u s s y ▁ ▁ ⓕ J e a n ▁ ▁ ⓑ 4 . 1 1 . 1 4
ⓢ M a r i n ▁ ▁ ⓕ M a r c e l ▁ ▁ ⓑ 1 0 . 8 . 0 6
ⓢ A m i c a l ▁ ▁ ⓕ E l o i ▁ ▁ ⓑ 1 1 . 1 0 . 0 4
ⓢ B i r o s ▁ ▁ ⓕ M a e l ▁ ▁ ⓑ 3 0 . 1 0 . 1 0"""

    expected_word_language_corpus = """ⓢ Caillet ▁ ⓕ Maurice ▁ ⓑ 28 ▁ . ▁ 9 ▁ . ▁ 06
ⓢ Reboul ▁ ⓕ Jean ▁ ⓑ 30 ▁ . ▁ 9 ▁ . ▁ 02
ⓢ Bareyre ▁ ⓕ Jean ▁ ⓑ 28 ▁ . ▁ 3 ▁ . ▁ 11
ⓢ Roussy ▁ ⓕ Jean ▁ ⓑ 4 ▁ . ▁ 11 ▁ . ▁ 14
ⓢ Marin ▁ ⓕ Marcel ▁ ⓑ 10 ▁ . ▁ 8 ▁ . ▁ 06
ⓢ Amical ▁ ⓕ Eloi ▁ ⓑ 11 ▁ . ▁ 10 ▁ . ▁ 04
ⓢ Biros ▁ ⓕ Mael ▁ ⓑ 30 ▁ . ▁ 10 ▁ . ▁ 10"""

    # Transcriptions with worker version are in lowercase
    if transcription_entities_worker_version:
        expected_char_language_corpus = expected_char_language_corpus.lower()
        expected_word_language_corpus = expected_word_language_corpus.lower()
        expected_subword_language_corpus = expected_subword_language_corpus.lower()

    # If we do not load entities, remove tokens
    if not load_entities:
        token_translations = {f"{token} ": "" for token in tokens}
        expected_char_language_corpus = ENTITY_TOKEN_SPACE.sub(
            "", expected_char_language_corpus
        )
        expected_word_language_corpus = ENTITY_TOKEN_SPACE.sub(
            "", expected_word_language_corpus
        )
        expected_subword_language_corpus = ENTITY_TOKEN_SPACE.sub(
            "", expected_subword_language_corpus
        )
    # Replace double spaces with regular space
    if not keep_spaces:
        expected_char_language_corpus = TWO_SPACES_LM_REGEX.sub(
            "▁", expected_char_language_corpus
        )
        expected_word_language_corpus = TWO_SPACES_LM_REGEX.sub(
            "▁", expected_word_language_corpus
        )
        expected_subword_language_corpus = TWO_SPACES_LM_REGEX.sub(
            "▁", expected_subword_language_corpus
        )

    assert (
        output / "language_model" / "corpus_characters.txt"
    ).read_text() == expected_char_language_corpus

    assert (
        output / "language_model" / "corpus_words.txt"
    ).read_text() == expected_word_language_corpus

    assert (
        output / "language_model" / "corpus_subwords.txt"
    ).read_text() == expected_subword_language_corpus

    # Check "language_tokens.txt"
    expected_language_tokens = [
        "▁" if t.isspace() else t for t in sorted(list(expected_charset))
    ]
    expected_language_tokens.append("◌")
    assert (output / "language_model" / "tokens.txt").read_text() == "\n".join(
        expected_language_tokens
    )

    # Check "language_lexicon.txt"
    expected_language_char_lexicon = [f"{t} {t}" for t in expected_language_tokens]
    assert (
        output / "language_model" / "lexicon_characters.txt"
    ).read_text() == "\n".join(expected_language_char_lexicon)

    word_vocab = set([word for word in expected_word_language_corpus.split()])
    expected_language_word_lexicon = [
        f"{word} {' '.join(word)}" for word in sorted(word_vocab)
    ]
    assert (output / "language_model" / "lexicon_words.txt").read_text() == "\n".join(
        expected_language_word_lexicon
    )

    subword_vocab = set(
        [subword for subword in expected_subword_language_corpus.split()]
    )
    expected_language_subword_lexicon = [
        f"{subword} {' '.join(subword)}" for subword in sorted(subword_vocab)
    ]
    assert (
        output / "language_model" / "lexicon_subwords.txt"
    ).read_text() == "\n".join(expected_language_subword_lexicon)


@pytest.mark.parametrize("allow_empty", (True, False))
def test_empty_transcription(allow_empty, mock_database):
    extractor = ArkindexExtractor(
        element_type=["text_line"],
        entity_separators=None,
        allow_empty=allow_empty,
    )
    element_no_transcription = Element(id="unknown")
    if allow_empty:
        assert extractor.extract_transcription(element_no_transcription) == ""
    else:
        with pytest.raises(NoTranscriptionError):
            extractor.extract_transcription(element_no_transcription)


@pytest.mark.parametrize("tokens", (None, EXTRACTION_DATA_PATH / "tokens.yml"))
def test_extract_transcription_no_translation(mock_database, tokens):
    extractor = ArkindexExtractor(
        element_type=["text_line"],
        entity_separators=None,
        tokens=tokens,
    )

    element = Element.get_by_id("test-page_1-line_1")
    # Deleting one of the two transcriptions from the element
    Transcription.get(
        Transcription.element == element,
        Transcription.worker_version_id == "worker_version_id",
    ).delete_instance(recursive=True)

    # Deleting all entities on the element remaining transcription while leaving the transcription intact
    if tokens:
        TranscriptionEntity.delete().where(
            TranscriptionEntity.transcription
            == Transcription.select().where(Transcription.element == element).get()
        ).execute()

    # Early return with only the element transcription text instead of a translation
    assert extractor.extract_transcription(element) == "Coupez  Bouis  7.12.14"


@pytest.mark.parametrize(
    "nestation, xml_output, separators",
    (
        # Non-nested
        (
            "non-nested-id",
            "<root>The <adj>great</adj> king <name>Charles</name> III has eaten \nwith <person>us</person>.</root>",
            None,
        ),
        # Non-nested no text between entities
        (
            "non-nested-id",
            "<root><adj>great</adj> <name>Charles</name>\n<person>us</person></root>",
            ["\n", " "],
        ),
        # Nested
        (
            "nested-id",
            "<root>The great king <fullname><name>Charles</name> III</fullname> has eaten \nwith <person>us</person>.</root>",
            None,
        ),
        # Nested no text between entities
        (
            "nested-id",
            "<root><fullname><name>Charles</name> III</fullname>\n<person>us</person></root>",
            ["\n", " "],
        ),
        # Special characters in entities
        (
            "special-chars-id",
            "<root>The <Arkindex_s_entity>great</Arkindex_s_entity> king <_Name_1_>Charles</_Name_1_> III has eaten \nwith <Person_>us</Person_>.</root>",
            None,
        ),
    ),
)
def test_entities_to_xml(mock_database, nestation, xml_output, separators):
    transcription = Transcription.get_by_id("tr-with-entities")
    assert (
        entities_to_xml(
            text=transcription.text,
            predictions=get_transcription_entities(
                transcription_id="tr-with-entities",
                entity_worker_versions=[nestation],
                supported_types=[
                    "name",
                    "fullname",
                    "person",
                    "adj",
                    "Arkindex's entity",
                    '"Name" (1)',
                    "Person /!\\",
                ],
            ),
            entity_separators=separators,
        )
        == xml_output
    )


@pytest.mark.parametrize(
    "supported_entities, xml_output, separators",
    (
        # <adj> missing, no text between entities
        (
            ["name", "person"],
            "<root><name>Charles</name>\n<person>us</person></root>",
            ["\n", " "],
        ),
        # <adj> missing, text between entities
        (
            ["name", "person"],
            "<root>The great king <name>Charles</name> III has eaten \nwith <person>us</person>.</root>",
            None,
        ),
    ),
)
def test_entities_to_xml_partial_entities(
    mock_database, supported_entities, xml_output, separators
):
    transcription = Transcription.get_by_id("tr-with-entities")
    assert (
        entities_to_xml(
            text=transcription.text,
            predictions=get_transcription_entities(
                transcription_id="tr-with-entities",
                entity_worker_versions=["non-nested-id"],
                supported_types=supported_entities,
            ),
            entity_separators=separators,
        )
        == xml_output
    )


@pytest.mark.parametrize(
    "transcription",
    (
        "Something\n",
        "Something\r",
        "Something\t",
        'Something"',
        "Something'",
        "Something<",
        "Something>",
        "Something&",
    ),
)
def test_entities_to_xml_no_encode(transcription):
    assert (
        entities_to_xml(
            text=transcription,
            # Empty queryset
            predictions=TranscriptionEntity.select().where(TranscriptionEntity.id == 0),
            entity_separators=None,
        )
        == f"<root>{transcription}</root>"
    )



# Copyright Teklia (contact@teklia.com) & Denis Coquenet
# This code is licensed under CeCILL-C

# -*- coding: utf-8 -*-
import logging

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s %(levelname)s/%(name)s: %(message)s",
)



# Copyright Teklia (contact@teklia.com) & Denis Coquenet
# This code is licensed under CeCILL-C

# -*- coding: utf-8 -*-
import json
import logging
import random
from copy import deepcopy

import numpy as np
import torch
import torch.multiprocessing as mp

from dan.ocr.manager.training import Manager
from dan.ocr.mlflow import MLFLOW_AVAILABLE
from dan.ocr.utils import update_config
from dan.utils import MLflowNotInstalled

if MLFLOW_AVAILABLE:
    import mlflow

    from dan.ocr.mlflow import make_mlflow_request, start_mlflow_run


logger = logging.getLogger(__name__)


def train(rank, params, mlflow_logging=False):
    torch.manual_seed(0)
    torch.cuda.manual_seed(0)
    np.random.seed(0)
    random.seed(0)
    torch.backends.cudnn.benchmark = False
    torch.backends.cudnn.deterministic = True

    params["training"]["device"]["ddp_rank"] = rank
    model = Manager(params)
    model.load_model()

    if params["dataset"]["tokens"] is not None:
        if "ner" not in params["training"]["metrics"]["train"]:
            params["training"]["metrics"]["train"].append("ner")
        if "ner" not in params["training"]["metrics"]["eval"]:
            params["training"]["metrics"]["eval"].append("ner")

    if mlflow_logging:
        logger.info("MLflow logging enabled")

    model.train(mlflow_logging=mlflow_logging)


def serialize_config(config):
    """
    Make every field of the configuration JSON-Serializable and remove sensitive information.

    - Classes are transformed using their name attribute
    - Functions are casted to strings
    """
    # Create a copy of the original config without erase it
    serialized_config = deepcopy(config)

    # Remove credentials to the config
    serialized_config["mlflow"]["s3_endpoint_url"] = ""
    serialized_config["mlflow"]["tracking_uri"] = ""
    serialized_config["mlflow"]["aws_access_key_id"] = ""
    serialized_config["mlflow"]["aws_secret_access_key"] = ""

    # Get the name of the class
    serialized_config["model"]["models"]["encoder"] = serialized_config["model"][
        "models"
    ]["encoder"].__name__
    serialized_config["model"]["models"]["decoder"] = serialized_config["model"][
        "models"
    ]["decoder"].__name__
    serialized_config["training"]["optimizers"]["all"]["class"] = serialized_config[
        "training"
    ]["optimizers"]["all"]["class"].__name__

    # Cast the functions to str
    serialized_config["dataset"]["config"]["augmentation"] = str(
        serialized_config["dataset"]["config"]["augmentation"]
    )
    serialized_config["training"]["nb_gpu"] = str(
        serialized_config["training"]["nb_gpu"]
    )

    return serialized_config


def start_training(config, mlflow_logging: bool) -> None:
    if (
        config["training"]["device"]["use_ddp"]
        and config["training"]["device"]["force"] in [None, "cuda"]
        and torch.cuda.is_available()
    ):
        mp.spawn(
            train,
            args=(config, mlflow_logging),
            nprocs=config["training"]["device"]["nb_gpu"],
        )
    else:
        train(0, config, mlflow_logging)


def run(config: dict):
    """
    Main program, training a new model, using a valid configuration
    """
    names = list(config["dataset"]["datasets"].keys())
    # We should only have one dataset
    assert len(names) == 1, f"Found {len(names)} datasets but only one is expected"

    dataset_name = names.pop()
    update_config(config)

    if config.get("mlflow") and not MLFLOW_AVAILABLE:
        logger.error(
            "Cannot log to MLflow. Please install the `mlflow` extra requirements."
        )
        raise MLflowNotInstalled()

    if not config.get("mlflow"):
        start_training(config, mlflow_logging=False)
    else:
        labels_path = config["dataset"]["datasets"][dataset_name] / "labels.json"
        with start_mlflow_run(config["mlflow"]) as (run, created):
            if created:
                logger.info(f"Started MLflow run with ID ({run.info.run_id})")
            else:
                logger.info(f"Resumed MLflow run with ID ({run.info.run_id})")

            make_mlflow_request(
                mlflow_method=mlflow.set_tags, tags={"Dataset": dataset_name}
            )
            # Get the labels json file
            labels_artifact = json.loads(labels_path.read_text())

            # Log MLflow artifacts
            for artifact, filename in [
                (serialize_config(config), "config.json"),
                (labels_artifact, "labels.json"),
            ]:
                make_mlflow_request(
                    mlflow_method=mlflow.log_dict,
                    dictionary=artifact,
                    artifact_file=filename,
                )
            start_training(config, mlflow_logging=True)



# Copyright Teklia (contact@teklia.com) & Denis Coquenet
# This code is licensed under CeCILL-C

# -*- coding: utf-8 -*-
"""
Train a new DAN model.
"""


from dan.ocr.evaluate import add_evaluate_parser  # noqa
from dan.ocr.predict import add_predict_parser  # noqa
from dan.ocr.train import run
from dan.utils import read_json


def add_train_parser(subcommands) -> None:
    parser = subcommands.add_parser(
        "train",
        description=__doc__,
        help=__doc__,
    )

    parser.add_argument(
        "--config",
        type=read_json,
        required=True,
        help="Configuration file.",
    )

    parser.set_defaults(func=run)



# Copyright Teklia (contact@teklia.com) & Denis Coquenet
# This code is licensed under CeCILL-C

# -*- coding: utf-8 -*-

from typing import Dict, List

import numpy as np
import torch
from torch import relu, softmax
from torch.nn import Conv1d, Dropout, Embedding, LayerNorm, Linear, Module, ModuleList
from torch.nn.init import xavier_uniform_
from torchaudio.models.decoder import CTCHypothesis, ctc_decoder

from dan.utils import LMTokenMapping, read_txt


class PositionalEncoding1D(Module):
    def __init__(self, dim, len_max, device):
        super(PositionalEncoding1D, self).__init__()
        self.pe = torch.zeros((1, dim, len_max), device=device, requires_grad=False)

        div = torch.exp(
            -torch.arange(0.0, dim, 2) / dim * torch.log(torch.tensor(10000.0))
        ).unsqueeze(1)
        l_pos = torch.arange(0.0, len_max)
        self.pe[:, ::2, :] = torch.sin(l_pos * div).unsqueeze(0)
        self.pe[:, 1::2, :] = torch.cos(l_pos * div).unsqueeze(0)

    def forward(self, x, start):
        """
        Add 1D positional encoding to x
        x: (B, C, L)
        start: index for x[:,:, 0]
        """
        if isinstance(start, int):
            return x + self.pe[:, :, start : start + x.size(2)].to(x.device)
        else:
            for i in range(x.size(0)):
                x[i] = x[i] + self.pe[0, :, start[i] : start[i] + x.size(2)]
            return x


class PositionalEncoding2D(Module):
    def __init__(self, dim, h_max, w_max, device):
        super(PositionalEncoding2D, self).__init__()
        self.pe = torch.zeros(
            (1, dim, h_max, w_max), device=device, requires_grad=False
        )

        div = torch.exp(
            -torch.arange(0.0, dim // 2, 2) / dim * torch.log(torch.tensor(10000.0))
        ).unsqueeze(1)
        w_pos = torch.arange(0.0, w_max)
        h_pos = torch.arange(0.0, h_max)
        self.pe[:, : dim // 2 : 2, :, :] = (
            torch.sin(h_pos * div).unsqueeze(0).unsqueeze(3).repeat(1, 1, 1, w_max)
        )
        self.pe[:, 1 : dim // 2 : 2, :, :] = (
            torch.cos(h_pos * div).unsqueeze(0).unsqueeze(3).repeat(1, 1, 1, w_max)
        )
        self.pe[:, dim // 2 :: 2, :, :] = (
            torch.sin(w_pos * div).unsqueeze(0).unsqueeze(2).repeat(1, 1, h_max, 1)
        )
        self.pe[:, dim // 2 + 1 :: 2, :, :] = (
            torch.cos(w_pos * div).unsqueeze(0).unsqueeze(2).repeat(1, 1, h_max, 1)
        )

    def forward(self, x):
        """
        Add 2D positional encoding to x
        x: (B, C, H, W)
        """
        return x + self.pe[:, :, : x.size(2), : x.size(3)]


class CustomMultiHeadAttention(Module):
    """
    Re-implementation of Multi-head Attention
    """

    def __init__(self, embed_dim, num_heads, dropout=0, proj_value=True):
        super().__init__()

        self.proj_value = proj_value

        self.in_proj_q = Linear(embed_dim, embed_dim)
        self.in_proj_k = Linear(embed_dim, embed_dim)
        if self.proj_value:
            self.in_proj_v = Linear(embed_dim, embed_dim)
        self.out_proj = Linear(embed_dim, embed_dim)

        self.num_heads = num_heads
        self.head_dim = embed_dim // num_heads
        self.scale_factor = float(self.head_dim) ** -0.5
        self.dropout = Dropout(dropout)

    def forward(
        self,
        query,
        key,
        value,
        key_padding_mask=None,
        attn_mask=None,
        output_weights=True,
    ):
        target_len, b, c = query.size()
        source_len = key.size(0)
        q = self.in_proj_q(query)
        k = self.in_proj_k(key)
        v = self.in_proj_v(value) if self.proj_value else value
        q = q * self.scale_factor

        q = torch.reshape(q, (target_len, b * self.num_heads, self.head_dim)).transpose(
            0, 1
        )
        k = torch.reshape(k, (source_len, b * self.num_heads, self.head_dim)).transpose(
            0, 1
        )
        v = torch.reshape(v, (source_len, b * self.num_heads, self.head_dim)).transpose(
            0, 1
        )

        attn_output_weights = torch.bmm(q, k.transpose(1, 2))

        if attn_mask is not None:
            attn_mask = attn_mask.unsqueeze(0)
            if attn_mask.dtype == torch.bool:
                attn_output_weights.masked_fill_(attn_mask, float("-inf"))
            else:
                attn_output_weights += attn_mask

        if key_padding_mask is not None:
            attn_output_weights = attn_output_weights.view(
                b, self.num_heads, target_len, source_len
            )

            attn_output_weights = attn_output_weights.masked_fill(
                key_padding_mask.unsqueeze(1).unsqueeze(2),
                float("-inf"),
            )
            attn_output_weights = attn_output_weights.view(
                b * self.num_heads, target_len, source_len
            )

        attn_output_weights_raw = softmax(attn_output_weights, dim=-1)

        attn_output_weights = self.dropout(attn_output_weights_raw)

        attn_output = torch.bmm(attn_output_weights, v)
        attn_output = attn_output.transpose(0, 1).contiguous().view(target_len, b, c)
        attn_output = self.out_proj(attn_output)

        if output_weights:
            attn_output_weights_raw = attn_output_weights_raw.view(
                b, self.num_heads, target_len, source_len
            )
            return attn_output, attn_output_weights_raw.sum(dim=1) / self.num_heads
        return attn_output

    def init_weights(self):
        xavier_uniform_(self.in_proj_q.weight)
        xavier_uniform_(self.in_proj_k.weight)
        if self.proj_value:
            xavier_uniform_(self.in_proj_v.weight)


class GlobalDecoderLayer(Module):
    """
    Transformer Decoder Layer
    """

    def __init__(self, params):
        super(GlobalDecoderLayer, self).__init__()
        self.self_att = CustomMultiHeadAttention(
            embed_dim=params["enc_dim"],
            num_heads=params["dec_num_heads"],
            proj_value=True,
            dropout=params["dec_att_dropout"],
        )

        self.norm1 = LayerNorm(params["enc_dim"])
        self.att = CustomMultiHeadAttention(
            embed_dim=params["enc_dim"],
            num_heads=params["dec_num_heads"],
            proj_value=True,
            dropout=params["dec_att_dropout"],
        )

        self.linear1 = Linear(params["enc_dim"], params["dec_dim_feedforward"])
        self.linear2 = Linear(params["dec_dim_feedforward"], params["enc_dim"])

        self.dropout = Dropout(params["dec_res_dropout"])

        self.norm2 = LayerNorm(params["enc_dim"])
        self.norm3 = LayerNorm(params["enc_dim"])

    def forward(
        self,
        tgt,
        memory_key,
        memory_value=None,
        tgt_mask=None,
        memory_mask=None,
        tgt_key_padding_mask=None,
        memory_key_padding_mask=None,
        predict_last_n_only=None,
    ):
        if memory_value is None:
            memory_value = memory_key

        self_att_query = tgt[-predict_last_n_only:] if predict_last_n_only else tgt

        tgt2, weights_self = self.self_att(
            self_att_query,
            tgt,
            tgt,
            attn_mask=tgt_mask,
            key_padding_mask=tgt_key_padding_mask,
            output_weights=True,
        )
        tgt = self_att_query + self.dropout(tgt2)
        tgt = self.norm1(tgt)
        att_query = tgt

        tgt2, weights = self.att(
            att_query,
            memory_key,
            memory_value,
            attn_mask=memory_mask,
            key_padding_mask=memory_key_padding_mask,
            output_weights=True,
        )

        tgt = att_query + self.dropout(tgt2)
        tgt = self.norm2(tgt)
        tgt2 = self.linear2(self.dropout(relu(self.linear1(tgt))))
        tgt = tgt + self.dropout(tgt2)
        tgt = self.norm3(tgt)
        return tgt, weights, weights_self


class GlobalAttDecoder(Module):
    """
    Stack of transformer decoder layers
    """

    def __init__(self, params):
        super(GlobalAttDecoder, self).__init__()

        self.decoder_layers = ModuleList(
            [GlobalDecoderLayer(params) for _ in range(params["dec_num_layers"])]
        )

    def forward(
        self,
        tgt,
        memory_key,
        memory_value,
        tgt_mask,
        memory_mask,
        tgt_key_padding_mask,
        memory_key_padding_mask,
        use_cache=False,
        cache=None,
        predict_last_n_only=False,
        keep_all_weights=False,
    ):
        output = tgt
        cache_t = list()
        all_weights = {"self": list(), "mix": list()}

        for i, dec_layer in enumerate(self.decoder_layers):
            output, weights, weights_self = dec_layer(
                output,
                memory_key=memory_key,
                memory_value=memory_value,
                tgt_mask=tgt_mask,
                memory_mask=memory_mask,
                tgt_key_padding_mask=tgt_key_padding_mask,
                memory_key_padding_mask=memory_key_padding_mask,
                predict_last_n_only=predict_last_n_only,
            )
            if use_cache:
                cache_t.append(output)
                if cache is not None:
                    output = torch.cat([cache[i], output], dim=0)
            if keep_all_weights:
                all_weights["self"].append(weights_self)
                all_weights["mix"].append(weights)
        if use_cache:
            cache = (
                torch.cat([cache, torch.stack(cache_t, dim=0)], dim=1)
                if cache is not None
                else torch.stack(cache_t, dim=0)
            )

        if predict_last_n_only:
            output = output[-predict_last_n_only:]

        if keep_all_weights:
            return output, all_weights, cache

        return output, weights, cache


class FeaturesUpdater(Module):
    """
    Module that handle 2D positional encoding
    """

    def __init__(self, params):
        super(FeaturesUpdater, self).__init__()
        self.pe_2d = PositionalEncoding2D(
            params["enc_dim"], params["h_max"], params["w_max"], params["device"]
        )

    def get_pos_features(self, features):
        return self.pe_2d(features)


class GlobalHTADecoder(Module):
    """
    DAN decoder module
    """

    def __init__(self, params):
        super(GlobalHTADecoder, self).__init__()
        self.dropout = Dropout(params["dec_pred_dropout"])
        self.dec_att_win = (
            params["attention_win"] if params["attention_win"] is not None else 1
        )

        self.features_updater = FeaturesUpdater(params)
        self.att_decoder = GlobalAttDecoder(params)

        self.emb = Embedding(
            num_embeddings=params["vocab_size"] + 3, embedding_dim=params["enc_dim"]
        )
        self.pe_1d = PositionalEncoding1D(
            params["enc_dim"], params["l_max"], params["device"]
        )

        vocab_size = params["vocab_size"] + 1
        self.end_conv = Conv1d(params["enc_dim"], vocab_size, kernel_size=1)

    def forward(
        self,
        features_1d,
        tokens,
        reduced_size,
        token_len,
        features_size,
        start=0,
        hidden_predict=None,
        cache=None,
        num_pred=None,
        keep_all_weights=False,
    ):
        device = features_1d.device

        # Token to Embedding
        pos_tokens = self.emb(tokens).permute(0, 2, 1)

        # Add 1D Positional Encoding
        pos_tokens = self.pe_1d(pos_tokens, start=start).permute(2, 0, 1)

        if num_pred is None:
            num_pred = tokens.size(1)

        # Use cache values to avoid useless computation at eval time
        if self.dec_att_win > 1 and cache is not None:
            cache = cache[:, -self.dec_att_win + 1 :]
        else:
            cache = None
        num_tokens_to_keep = min(
            [num_pred + self.dec_att_win - 1, pos_tokens.size(0), token_len[0]]
        )
        pos_tokens = pos_tokens[-num_tokens_to_keep:]

        # Generate dynamic masks
        target_mask = self.generate_target_mask(
            tokens.size(1), device
        )  # Use only already predicted tokens (causal)
        memory_mask = None  # Use all feature position

        # Generate static masks
        key_target_mask = self.generate_token_mask(
            token_len, tokens.size(), device
        )  # Use all token except padding
        key_memory_mask = self.generate_enc_mask(
            reduced_size, features_size, device
        )  # Use all feature position except padding

        target_mask = target_mask[-num_pred:, -num_tokens_to_keep:]
        key_target_mask = key_target_mask[:, -num_tokens_to_keep:]

        output, weights, cache = self.att_decoder(
            pos_tokens,
            memory_key=features_1d,
            memory_value=features_1d,
            tgt_mask=target_mask,
            memory_mask=memory_mask,
            tgt_key_padding_mask=key_target_mask,
            memory_key_padding_mask=key_memory_mask,
            use_cache=True,
            cache=cache,
            predict_last_n_only=num_pred,
            keep_all_weights=keep_all_weights,
        )

        dp_output = self.dropout(relu(output))
        preds = self.end_conv(dp_output.permute(1, 2, 0))

        if not keep_all_weights:
            weights = torch.sum(weights, dim=1, keepdim=True).reshape(
                -1, 1, features_size[2], features_size[3]
            )
        return output, preds, hidden_predict, cache, weights

    def generate_enc_mask(self, batch_reduced_size, total_size, device):
        """
        Generate mask for encoded features
        """
        batch_size, _, h_max, w_max = total_size
        mask = torch.ones((batch_size, h_max, w_max), dtype=torch.bool, device=device)
        for i, (h, w) in enumerate(batch_reduced_size):
            mask[i, :h, :w] = False
        return torch.flatten(mask, start_dim=1, end_dim=2)

    def generate_token_mask(self, token_len, total_size, device):
        """
        Generate mask for tokens per sample
        """
        batch_size, len_max = total_size
        mask = torch.zeros((batch_size, len_max), dtype=torch.bool, device=device)
        for i, len_ in enumerate(token_len):
            mask[i, :len_] = False
        return mask

    def generate_target_mask(self, target_len, device):
        """
        Generate mask for tokens per time step (teacher forcing)
        """
        if self.dec_att_win == 1:
            return torch.triu(
                torch.ones((target_len, target_len), dtype=torch.bool, device=device),
                diagonal=1,
            )
        else:
            return torch.logical_not(
                torch.logical_and(
                    torch.tril(
                        torch.ones(
                            (target_len, target_len), dtype=torch.bool, device=device
                        ),
                        diagonal=0,
                    ),
                    torch.triu(
                        torch.ones(
                            (target_len, target_len), dtype=torch.bool, device=device
                        ),
                        diagonal=-self.dec_att_win + 1,
                    ),
                )
            )


class CTCLanguageDecoder:
    """
    Initialize a CTC decoder with n-gram language modeling.
    :param language_model_path: Path to a KenLM or ARPA language model.
    :param lexicon_path: Path to a lexicon file containing the possible words and corresponding spellings.
            Each line consists of a word and its space separated spelling. If `None`, uses lexicon-free decoding.
    :param tokens_path: Path to a file containing valid tokens. If using a file, the expected
            format is for tokens mapping to the same index to be on the same line.
    :param language_model_weight: Weight of the language model.
    :param temperature: Temperature for model calibreation.
    """

    def __init__(
        self,
        language_model_path: str,
        lexicon_path: str,
        tokens_path: str,
        language_model_weight: float = 1.0,
        temperature: float = 1.0,
    ):
        self.mapping = LMTokenMapping()
        self.language_model_weight = language_model_weight
        self.temperature = temperature
        self.tokens_to_index = {
            token: i for i, token in enumerate(read_txt(tokens_path).split("\n"))
        }
        self.index_to_token = {i: token for token, i in self.tokens_to_index.items()}
        self.blank_token_id = self.tokens_to_index[self.mapping.ctc.encoded]

        # Torchaudio's decoder
        # https://pytorch.org/audio/master/generated/torchaudio.models.decoder.ctc_decoder.html
        self.decoder = ctc_decoder(
            lm=language_model_path,
            lexicon=lexicon_path,
            tokens=tokens_path,
            lm_weight=self.language_model_weight,
            blank_token=self.mapping.ctc.encoded,
            sil_token=self.mapping.space.encoded,
            unk_word="⁇",
            nbest=1,
        )
        # No GPU support
        self.device = torch.device("cpu")

    def add_ctc_frames(
        self, batch_features: torch.FloatTensor, batch_frames: torch.LongTensor
    ) -> tuple[torch.FloatTensor, torch.LongTensor]:
        """
        Add CTC frames between each characters to avoid duplicate removal.
        """
        high_prob = batch_features.max()
        low_prob = batch_features.min()
        batch_size, n_frames, n_tokens = batch_features.shape
        # Reset probabilities for the CTC token
        batch_features[:, :, -1] = (
            torch.ones(
                (batch_size, n_frames),
                dtype=torch.float32,
                device=batch_features.device,
            )
            * low_prob
        )

        # Create a frame with high probability CTC token
        ctc_probs = (
            torch.ones(
                (batch_size, 1, n_tokens),
                dtype=torch.float32,
                device=batch_features.device,
            )
            * low_prob
        )
        ctc_probs[:, :, self.blank_token_id] = high_prob
        ctc_probs = ctc_probs

        # Insert the CTC frame between regular frames
        for fn in range(batch_frames.max() - 1):
            batch_features = torch.cat(
                [
                    batch_features[:, : 2 * fn + 1, :],
                    ctc_probs,
                    batch_features[:, 2 * fn + 1 :, :],
                ],
                dim=1,
            )

        # Update the number of frames
        batch_frames = 2 * batch_frames - 1
        return batch_features, batch_frames

    def post_process(
        self, hypotheses: List[CTCHypothesis], batch_sizes: torch.LongTensor
    ) -> Dict[str, List[str | float]]:
        """
        Post-process hypotheses to output JSON. Exports only the best hypothesis for each image.
        :param hypotheses: List of hypotheses returned by the decoder.
        :param batch_sizes: Prediction length of size batch_size.
        :return: A dictionary containing the hypotheses and their confidences.
        """
        out = {}
        # Replace <space> by an actual space and format string
        out["text"] = [
            "".join(
                [
                    self.mapping.display[self.index_to_token[token]]
                    if self.index_to_token[token] in self.mapping.display
                    else self.index_to_token[token]
                    for token in hypothesis[0].tokens.tolist()
                ]
            ).strip()
            for hypothesis in hypotheses
        ]
        # Normalize confidence score
        out["confidence"] = [
            np.around(
                np.exp(
                    hypothesis[0].score
                    / ((self.language_model_weight + 1) * length.item())
                ),
                2,
            )
            for hypothesis, length in zip(hypotheses, batch_sizes)
        ]
        return out

    def __call__(
        self, batch_features: torch.FloatTensor, batch_frames: torch.LongTensor
    ) -> Dict[str, List[str | float]]:
        """
        Decode a feature vector using n-gram language modelling.
        :param batch_features: Feature vector of size (batch_size, n_tokens, n_frames).
        :param batch_frames: Prediction length of size batch_size.
        :return: A dictionary containing the hypotheses and their confidences.
        """
        # Reshape from (batch_size, n_tokens, n_frames) to (batch_size, n_frames, n_tokens)
        batch_features = batch_features.permute((0, 2, 1))

        # Insert CTC frames to avoid getting rid of duplicates
        # Make sure that the CTC token has low probs for other frames
        batch_features, batch_frames = self.add_ctc_frames(batch_features, batch_frames)

        # Apply log softmax
        batch_features = torch.nn.functional.log_softmax(
            batch_features / self.temperature, dim=-1
        )

        # No GPU support for torchaudio's ctc_decoder
        batch_features = batch_features.to(self.device)
        batch_frames = batch_frames.to(self.device)

        # Decode
        hypotheses = self.decoder(batch_features, batch_frames)
        return self.post_process(hypotheses, batch_frames)



# Copyright Teklia (contact@teklia.com) & Denis Coquenet
# This code is licensed under CeCILL-C

# -*- coding: utf-8 -*-
import logging
import re
from enum import Enum
from typing import Dict, List, Tuple

import cv2
import matplotlib.pyplot as plt
import numpy as np
import torch
from PIL import Image
from torchvision.transforms.functional import to_pil_image

from dan.utils import EntityType

logger = logging.getLogger(__name__)


class Level(str, Enum):
    Char = "char"
    Word = "word"
    Line = "line"
    NER = "ner"

    def __str__(self):
        return self.value


def parse_delimiters(delimiters: List[str]) -> re.Pattern:
    return re.compile(rf"[^{'|'.join(delimiters)}]+")


def compute_offsets_by_level(full_text: str, level: Level, text_list: List[str]):
    """
    Compute and return the list of offset between each text part.
    :param full_text: predicted text.
    :param level: Level to use from [char, word, line, ner].
    :param text_list: list of text to use.
    Returns a list of offsets.
    """
    # offsets[idx] = number of characters between text_list[idx-1] and text_list[idx]
    offsets = [int(level != Level.Char)] * (len(text_list) - 1)

    if level == Level.NER:
        # Start after the first entity
        cursor = len(text_list[0])
        for idx, split in enumerate(text_list[1:]):
            # Number of characters between this entity and the previous one
            offsets[idx] = full_text[cursor:].index(split)
            cursor += offsets[idx] + len(split)

    # Last offset is not used, padded with a 0 to match the length of text_list
    return offsets + [0]


def compute_prob_by_separator(
    characters: str, probabilities: List[float], separator: re.Pattern
) -> Tuple[List[str], List[np.float64]]:
    """
    Split text and confidences using separators and return a list of average confidence scores.
    :param characters: list of characters.
    :param probabilities: list of character probabilities.
    :param separators: regex for separators. Use parse_delimiters(["\n", " "]) for word confidences and parse_delimiters(["\n"]) for line confidences.
    Returns a list confidence scores.
    """
    # match anything except separators, get start and end index
    matches = [(m.start(), m.end()) for m in separator.finditer(characters)]

    # Iterate over text pieces and compute mean confidence
    return [characters[start:end] for (start, end) in matches], [
        np.mean(probabilities[start:end]) for (start, end) in matches
    ]


def split_text(
    text: str,
    level: Level,
    char_separators: re.Pattern,
    word_separators: re.Pattern,
    line_separators: re.Pattern,
    tokens_separators: re.Pattern | None = None,
) -> Tuple[List[str], List[int]]:
    """
    Split text into a list of characters, word, or lines.
    :param text: Text prediction from DAN
    :param level: Level to visualize from [char, word, line, ner]
    :param char_separators: Pattern used to find tokens in the charset
    :param word_separators: Pattern used to find words
    :param line_separators: Pattern used to find lines
    :param tokens_separators: Pattern used to find NER entities
    """
    match level:
        case Level.Char:
            text_split = char_separators.findall(text)
        # split into words
        case Level.Word:
            text_split = word_separators.split(text)
        # split into lines
        case Level.Line:
            text_split = line_separators.split(text)
        # split into entities
        case Level.NER:
            if not tokens_separators:
                logger.error("Cannot compute NER level: tokens not found")
                return [], []
            text_split = tokens_separators.findall(text)
        case _:
            logger.error(f"Level should be either {list(map(str, Level))}")
            return [], []

    return text_split, compute_offsets_by_level(text, level, text_split)


def split_text_and_confidences(
    text: str,
    confidences: List[float],
    level: Level,
    char_separators: re.Pattern,
    word_separators: re.Pattern,
    line_separators: re.Pattern,
    tokens_separators: re.Pattern | None = None,
) -> Tuple[List[str], List[np.float64], List[int]]:
    """
    Split text into a list of characters, words or lines with corresponding confidences scores
    :param text: Text prediction from DAN
    :param confidences: Character confidences
    :param level: Level to visualize from [char, word, line, ner]
    :param char_separators: Pattern used to find tokens of the charset
    :param word_separators: Pattern used to find words
    :param line_separators: Pattern used to find lines
    :param tokens_separators: Pattern used to find NER entities
    """
    match level:
        case Level.Char:
            texts = char_separators.findall(text)
        case Level.Word:
            texts, confidences = compute_prob_by_separator(
                text, confidences, word_separators
            )
        case Level.Line:
            texts, confidences = compute_prob_by_separator(
                text, confidences, line_separators
            )
        case Level.NER:
            if not tokens_separators:
                logger.error("Cannot compute NER level: tokens not found")
                return [], [], []

            texts, confidences = compute_prob_by_separator(
                text, confidences, tokens_separators
            )
        case _:
            logger.error(f"Level should be either {list(map(str, Level))}")
            return [], [], []

    return (
        texts,
        [np.around(num, 2) for num in confidences],
        compute_offsets_by_level(text, level, texts),
    )


def get_predicted_polygons_with_confidence(
    text: str,
    weights: np.ndarray,
    confidences: List[float],
    level: Level,
    height: int,
    width: int,
    char_separators: re.Pattern,
    max_object_height: int = 50,
    word_separators: re.Pattern = parse_delimiters(["\n", " "]),
    line_separators: re.Pattern = parse_delimiters(["\n"]),
    tokens_separators: re.Pattern | None = None,
) -> List[dict]:
    """
    Returns the polygons of each object of the current prediction
    :param text: Text predicted by DAN
    :param weights: Attention weights of size (n_char, feature_height, feature_width)
    :param confidences: Character confidences
    :param level: Level to display (must be in [char, word, line, ner])
    :param height: Original image height
    :param width: Original image width
    :param char_separators: Pattern used to find tokens of the charset
    :param max_object_height: Maximum height of predicted objects.
    :param word_separators: Pattern used to find words
    :param line_separators: Pattern used to find lines
    :param tokens_separators: Pattern used to find NER entities
    """
    # Split text into characters, words or lines
    text_list, confidence_list, offsets = split_text_and_confidences(
        text,
        confidences,
        level,
        char_separators,
        word_separators,
        line_separators,
        tokens_separators,
    )

    max_value = weights.sum(0).max()
    polygons = []
    start_index = 0
    for text_piece, confidence, offset in zip(text_list, confidence_list, offsets):
        polygon, _ = get_polygon(
            text_piece,
            max_value,
            start_index,
            weights,
            max_object_height=max_object_height,
            size=(width, height),
        )
        start_index += len(text_piece) + offset
        if not polygon:
            continue
        polygon["text"] = text_piece
        polygon["text_confidence"] = confidence
        polygons.append(polygon)
    return polygons


def compute_coverage(
    text: str,
    max_value: np.float32,
    offset: int,
    attentions: np.ndarray,
    size: Tuple[int, int],
) -> np.ndarray:
    """
    Aggregates attention maps for the current text piece (char, word, line)
    :param text: Text piece selected with offset after splitting DAN prediction
    :param max_value: Maximum "attention intensity" for parts of a text piece, used for normalization
    :param offset: Offset value to get the relevant part of text piece
    :param attentions: Attention weights of size (n_char, feature_height, feature_width)
    :param size: Target size (width, height) to resize the coverage vector
    """
    _, height, width = attentions.shape

    # blank vector to accumulate weights for the current text
    coverage_vector = np.zeros((height, width))
    for i in range(len(text)):
        local_weight = cv2.resize(attentions[i + offset], (width, height))
        coverage_vector = np.clip(coverage_vector + local_weight, 0, 1)

    # Normalize coverage vector
    coverage_vector = (coverage_vector / max_value * 255).astype(np.uint8)

    # Resize it
    if size:
        coverage_vector = cv2.resize(coverage_vector, size)

    return coverage_vector


def blend_coverage(
    coverage_vector: np.ndarray,
    image: Image.Image,
    scale: float,
    alpha_factor: float,
    color_map: str,
) -> Image.Image:
    """
    Blends current coverage_vector over original image, used to make an attention map.
    :param coverage_vector: Aggregated attention weights of the current text piece, resized to image. size: (n_char, image_height, image_width)
    :param image: Input image in PIL format
    :param scale: Scaling factor for the output gif image
    :param alpha_factor: Alpha factor that controls how much the attention map is shown to the user during prediction. (higher value means more transparency for the attention map, commonly between 0.5 and 1.0)
    :param color_map: Colormap to use for the attention map (from matplotlib colormaps)
    """
    height, width = coverage_vector.shape

    # Make a colormap so colors are can be applied based on a scale
    cmap = plt.get_cmap(color_map)

    # Normalize the coverage_vector to the range [0, 255]
    norm = plt.Normalize(vmin=0, vmax=255)

    # Apply the colormap to the normalized coverage_vector to get color values
    color = cmap(norm(coverage_vector))

    # Get a transparency map from the coverage_vector with a power function
    # higher alpha_factor means more transparency of the attention map
    #  with 1 as no change (commonly between 0.5 and 1.0)
    alpha = (coverage_vector / 255.0) ** alpha_factor

    # Add the alpha channel to the color map
    color = np.dstack((color[:, :, :3], alpha))

    # Convert the color map to a RGBA image and all channels to a 0-255 range (as all values are between 0 and 1)
    color = Image.fromarray((color * 255).astype(np.uint8), "RGBA")

    # Convert the original image to RGBA
    image = image.convert("RGBA")

    # Blend the two images together so the original image shows through the attention map
    blend = Image.alpha_composite(image, color)

    # Resize the image to the desired scale in order to reduce the time need to create the gif
    blend = blend.resize((int(width * scale), int(height * scale)), Image.LANCZOS)

    return blend


def compute_contour_metrics(
    coverage_vector: np.ndarray, contour: np.ndarray
) -> Tuple[np.float64, np.float64]:
    """
    Compute the contours's area and the mean value inside it.
    :param coverage_vector: Aggregated attention weights of the current text piece, resized to image. size: (n_char, image_height, image_width)
    :param contour: Contour of the current attention blob
    """
    # draw the contour zone
    mask = np.zeros(coverage_vector.shape, dtype=np.uint8)
    cv2.drawContours(mask, [contour], -1, (255), -1)

    max_value = np.where(mask > 0, coverage_vector, 0).max() / 255
    area = cv2.contourArea(contour)
    return max_value, max_value * area


def polygon_to_bbx(polygon: np.ndarray) -> List[Tuple[int, int]]:
    x, y, w, h = cv2.boundingRect(polygon)
    return [[x, y], [x + w, y], [x + w, y + h], [x, y + h]]


def threshold(mask: np.ndarray) -> np.ndarray:
    """
    Threshold a grayscale mask.
    :param mask: a grayscale image (np.array)
    """
    min_kernel = 1
    max_kernel = mask.shape[1] // 100

    # Blur and apply Otsu thresholding
    blur = cv2.GaussianBlur(mask, (15, 15), 0)
    _, bin_mask = cv2.threshold(blur, 0, 255, cv2.THRESH_BINARY + cv2.THRESH_OTSU)
    # Apply dilation
    kernel_width = cv2.getStructuringElement(cv2.MORPH_CROSS, (max_kernel, min_kernel))
    dilated = cv2.dilate(bin_mask, kernel_width, iterations=3)
    return np.asarray(dilated, dtype=np.uint8)


def get_polygon(
    text: str,
    max_value: np.float32,
    offset: int,
    weights: np.ndarray,
    size: Tuple[int, int] | None = None,
    max_object_height: int = 50,
) -> Tuple[dict, np.ndarray | None]:
    """
    Gets polygon associated with element of current text_piece, indexed by offset
    :param text: Text piece selected with offset after splitting DAN prediction
    :param max_value: Maximum "attention intensity" for parts of a text piece, used for normalization
    :param offset: Offset value to get the relevant part of text piece
    :param size: Target size (width, height) to resize the coverage vector
    :param max_object_height: Maximum height of predicted objects.
    """
    # Compute coverage vector
    coverage_vector = compute_coverage(text, max_value, offset, weights, size=size)

    # Generate a binary image for the current channel.
    bin_mask = threshold(coverage_vector)

    coord, confidence = (
        get_grid_search_contour(coverage_vector, bin_mask, height=max_object_height)
        if max_object_height
        else get_best_contour(coverage_vector, bin_mask)
    )
    if not coord or confidence is None:
        return {}, None

    # Format for JSON
    polygon = {
        "confidence": confidence,
        "polygon": coord,
    }
    simplified_contour = np.expand_dims(np.array(coord, dtype=np.int32), axis=1)

    return polygon, simplified_contour


def get_best_contour(coverage_vector, bin_mask):
    """
    Detect the objects contours using opencv and select the best one.
    """
    contours, _ = cv2.findContours(bin_mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)

    if not contours:
        return [], None

    # Select best contour
    metrics = [compute_contour_metrics(coverage_vector, cnt) for cnt in contours]
    confidences, scores = map(list, zip(*metrics))
    best_contour = contours[np.argmax(scores)]
    coord = polygon_to_bbx(np.squeeze(best_contour))
    confidence = round(confidences[np.argmax(scores)], 2)
    return coord, confidence


def get_grid_search_contour(coverage_vector, bin_mask, height=50):
    """
    Perform grid search to find the best contour with fixed width and height.
    """
    # Limit search area based on attention values
    roi = np.argwhere(bin_mask == 255)

    if not np.any(roi):
        return [], None

    y_min, y_max = roi[:, 0].min(), roi[:, 0].max()

    # Limit bounding box shape
    width = bin_mask.shape[1]

    best_sum_att = 0
    for y in range(y_min, max(y_max - height, y_min + 1)):
        sum_att = coverage_vector[y : y + height, 0:width].sum()
        if sum_att > best_sum_att:
            best_sum_att = sum_att
            best_coordinates = [0, y, width, height]
            confidence = coverage_vector[y : y + height, 0:width].max()

    # Format for JSON
    x, y, w, h = best_coordinates
    coord = [[x, y], [x + w, y], [x + w, y + h], [x, y + h]]
    confidence = round(confidence / 255.0, 2)
    return coord, confidence


def plot_attention(
    image: torch.Tensor,
    text: str,
    weights: np.ndarray,
    level: Level,
    scale: float,
    outname: str,
    alpha_factor: float,
    color_map: str,
    max_object_height: int = 50,
    word_separators: re.Pattern = parse_delimiters(["\n", " "]),
    line_separators: re.Pattern = parse_delimiters(["\n"]),
    tokens: Dict[str, EntityType] = {},
    display_polygons: bool = False,
) -> None:
    """
    Create a gif by blending attention maps to the image for each text piece (char, word or line)
    :param image: Input image as torch.Tensor
    :param text: Text predicted by DAN
    :param weights: Attention weights of size (n_char, feature_height, feature_width)
    :param level: Level to display (must be in [char, word, line, ner])
    :param scale: Scaling factor for the output gif image
    :param outname: Name of the gif image
    :param alpha_factor: Alpha factor that controls how much the attention map is shown to the user during prediction. (higher value means more transparency for the attention map, commonly between 0.5 and 1.0)
    :param color_map: Colormap to use for the attention map
    :param max_object_height: Maximum height of predicted objects.
    :param word_separators: List of word separators
    :param line_separators: List of line separators
    :param tokens: NER tokens used
    :param display_polygons: Whether to plot extracted polygons
    """
    image = to_pil_image(image)
    attention_map = []

    # Split text into characters, words or lines
    text_list, offsets = split_text(
        text, level, word_separators, line_separators, tokens
    )

    # Iterate on characters, words or lines
    tot_len = 0
    max_value = weights.sum(0).max()

    for text_piece, offset in zip(text_list, offsets):
        # Accumulate weights for the current word/line and resize to original image size
        coverage_vector = compute_coverage(
            text_piece, max_value, tot_len, weights, (image.width, image.height)
        )

        # Blend coverage vector with original image to make an attention map
        blended = blend_coverage(coverage_vector, image, scale, alpha_factor, color_map)

        # Get polygons if flag is set:
        if display_polygons:
            # Draw the contour
            _, contour = get_polygon(
                text_piece,
                max_value,
                tot_len,
                weights,
                max_object_height=max_object_height,
                size=(image.width, image.height),
            )

            if contour is not None:
                # The image has been scaled so we need to scale the contour
                contour = (contour * scale).astype(np.int32)

                # Draw the contour with a thickness based on the scale in red
                cv2.drawContours(
                    blended := np.array(blended),
                    [contour],
                    0,
                    (255, 0, 0, 1),
                    int(5 * scale),
                )

                # Make the np.array with drawn contours back into a PIL image
                blended = Image.fromarray(blended, "RGBA")

        # Keep track of text length
        tot_len += len(text_piece) + offset

        # Append the blended image to the list of attention maps to be used for the .gif
        attention_map.append(blended)

    if not attention_map:
        return

    attention_map[0].save(
        outname,
        save_all=True,
        format="GIF",
        append_images=attention_map[1:],
        duration=1000,
        loop=True,
    )



#!/usr/bin/env python
# Copyright Teklia (contact@teklia.com) & Denis Coquenet
# This code is licensed under CeCILL-C

# -*- coding: utf-8 -*-

from pathlib import Path
from typing import List

from setuptools import find_packages, setup


def parse_requirements_line(line) -> str:
    # Special case for git requirements
    if line.startswith("git+http"):
        assert "@" in line, "Branch should be specified with suffix (ex: @master)"
        assert (
            "#egg=" in line
        ), "Package name should be specified with suffix (ex: #egg=kraken)"
        package_name: str = line.split("#egg=")[-1]
        return f"{package_name} @ {line}"
    # Special case for submodule requirements
    elif line.startswith("-e"):
        package_path: str = line.split(" ")[-1]
        package = Path(package_path).resolve()
        return f"{package.name} @ file://{package}"
    else:
        return line


def parse_requirements(filename: str) -> List[str]:
    path = Path(__file__).parent.resolve() / filename
    assert path.exists(), f"Missing requirements: {path}"
    return list(
        map(parse_requirements_line, map(str.strip, path.read_text().splitlines()))
    )


setup(
    name="atr-dan",
    version=Path("VERSION").read_text(),
    description="Teklia DAN",
    long_description=Path("README.md").read_text(),
    long_description_content_type="text/markdown",
    author="Teklia",
    author_email="contact@teklia.com",
    url="https://gitlab.teklia.com/atr/dan",
    python_requires=">=3.10",
    install_requires=parse_requirements("requirements.txt"),
    packages=find_packages(),
    entry_points={
        "console_scripts": [
            "teklia-dan=dan.cli:main",
        ]
    },
    extras_require={
        "docs": parse_requirements("doc-requirements.txt"),
        "mlflow": parse_requirements("mlflow-requirements.txt"),
    },
    license="CeCILL-C",
    license_files=(
        "LICENSE",
        "LICENCE",
    ),
)



# Copyright Teklia (contact@teklia.com) & Denis Coquenet
# This code is licensed under CeCILL-C

# -*- coding: utf-8 -*-

from operator import itemgetter

import pytest

from arkindex_export import Dataset, DatasetElement, Element
from dan.datasets.extract.arkindex import TRAIN_NAME
from dan.datasets.extract.db import (
    get_dataset_elements,
    get_elements,
    get_transcription_entities,
    get_transcriptions,
)


def test_get_dataset_elements(mock_database):
    """
    Assert dataset elements retrieval output against verified results
    """
    dataset_elements = get_dataset_elements(
        dataset=Dataset.select().get(),
        split=TRAIN_NAME,
    )

    # ID verification
    assert all(
        isinstance(dataset_element, DatasetElement)
        for dataset_element in dataset_elements
    )
    assert [dataset_element.element.id for dataset_element in dataset_elements] == [
        "train-page_1",
        "train-page_2",
    ]


def test_get_elements(mock_database):
    """
    Assert elements retrieval output against verified results
    """
    elements = get_elements(
        parent_id="train-page_1",
        element_type=["text_line"],
    )

    # ID verification
    assert all(isinstance(element, Element) for element in elements)
    assert [element.id for element in elements] == [
        "train-page_1-line_1",
        "train-page_1-line_2",
        "train-page_1-line_3",
        "train-page_1-line_4",
    ]


@pytest.mark.parametrize(
    "worker_versions",
    ([False], ["worker_version_id"], [], [False, "worker_version_id"]),
)
def test_get_transcriptions(worker_versions, mock_database):
    """
    Assert transcriptions retrieval output against verified results
    """
    element_id = "train-page_1-line_1"
    transcriptions = get_transcriptions(
        element_id=element_id,
        transcription_worker_versions=worker_versions,
    )

    expected_transcriptions = []
    if not worker_versions or False in worker_versions:
        expected_transcriptions.append(
            {
                "text": "Caillet  Maurice  28.9.06",
                "worker_version_id": None,
            }
        )

    if not worker_versions or "worker_version_id" in worker_versions:
        expected_transcriptions.append(
            {
                "text": "caillet  maurice  28.9.06",
                "worker_version_id": "worker_version_id",
            }
        )

    assert (
        sorted(
            [
                {
                    "text": transcription.text,
                    "worker_version_id": transcription.worker_version.id
                    if transcription.worker_version
                    else None,
                }
                for transcription in transcriptions
            ],
            key=itemgetter("text"),
        )
        == expected_transcriptions
    )


@pytest.mark.parametrize("worker_version", (False, "worker_version_id", None))
@pytest.mark.parametrize(
    "supported_types", (["surname"], ["surname", "firstname", "birthdate"])
)
def test_get_transcription_entities(worker_version, mock_database, supported_types):
    transcription_id = "train-page_1-line_1" + (worker_version or "")
    entities = get_transcription_entities(
        transcription_id=transcription_id,
        entity_worker_versions=[worker_version],
        supported_types=supported_types,
    )

    expected_entities = [
        {
            "name": "Caillet",
            "type": "surname",
            "offset": 0,
            "length": 7,
        },
        {
            "name": "Maurice",
            "type": "firstname",
            "offset": 9,
            "length": 7,
        },
        {
            "name": "28.9.06",
            "type": "birthdate",
            "offset": 18,
            "length": 7,
        },
    ]

    expected_entities = list(
        filter(lambda ent: ent["type"] in supported_types, expected_entities)
    )
    for entity in expected_entities:
        if worker_version:
            entity["name"] = entity["name"].lower()
        entity["worker_version"] = worker_version or None

    assert (
        sorted(
            entities,
            key=itemgetter("offset"),
        )
        == expected_entities
    )



# Copyright Teklia (contact@teklia.com) & Denis Coquenet
# This code is licensed under CeCILL-C

# -*- coding: utf-8 -*-

from operator import attrgetter

import pytest
import torch
import yaml

from dan.ocr.train import train
from dan.ocr.utils import update_config
from tests.conftest import FIXTURES


@pytest.mark.parametrize(
    "expected_best_model_name, expected_last_model_name, params_res",
    (
        (
            "best_0.pt",
            "last_3.pt",
            {
                "parameters": {
                    "max_char_prediction": 30,
                    "encoder": {"dropout": 0.5},
                    "decoder": {
                        "enc_dim": 256,
                        "l_max": 15000,
                        "h_max": 500,
                        "w_max": 1000,
                        "dec_num_layers": 8,
                        "dec_num_heads": 4,
                        "dec_res_dropout": 0.1,
                        "dec_pred_dropout": 0.1,
                        "dec_att_dropout": 0.1,
                        "dec_dim_feedforward": 256,
                        "vocab_size": 96,
                        "attention_win": 100,
                    },
                    "preprocessings": [
                        {
                            "max_height": 2000,
                            "max_width": 2000,
                            "type": "max_resize",
                        }
                    ],
                    "mean": [
                        242.10595854671013,
                        242.10595854671013,
                        242.10595854671013,
                    ],
                    "std": [28.29919517652322, 28.29919517652322, 28.29919517652322],
                },
            },
        ),
    ),
)
def test_train(
    expected_best_model_name,
    expected_last_model_name,
    params_res,
    training_config,
    tmp_path,
):
    update_config(training_config)

    # Use the tmp_path as base folder
    training_config["training"]["output_folder"] = (
        tmp_path / training_config["training"]["output_folder"]
    )

    train(0, training_config)

    # There should only be two checkpoints left
    checkpoints = (
        tmp_path / training_config["training"]["output_folder"] / "checkpoints"
    )

    # Make sure we only have these two checkpoints
    assert sorted(map(attrgetter("name"), checkpoints.iterdir())) == [
        expected_best_model_name,
        expected_last_model_name,
    ]

    # Check that the trained model is correct
    for model_name in [expected_best_model_name, expected_last_model_name]:
        expected_model = torch.load(FIXTURES / "training" / "models" / model_name)
        trained_model = torch.load(
            checkpoints / model_name,
        )

        # Check the optimizers parameters
        for trained, expected in zip(
            trained_model["optimizers_named_params"]["encoder"],
            expected_model["optimizers_named_params"]["encoder"],
        ):
            for (trained_param, trained_tensor), (
                expected_param,
                expected_tensor,
            ) in zip(trained.items(), expected.items()):
                assert trained_param == expected_param
                assert torch.allclose(
                    trained_tensor, expected_tensor, rtol=1e-05, atol=1e-03
                )

        # Check the optimizer encoder and decoder state dicts
        for optimizer_part in [
            "optimizer_encoder_state_dict",
            "optimizer_decoder_state_dict",
        ]:
            for trained, expected in zip(
                trained_model[optimizer_part]["state"].values(),
                expected_model[optimizer_part]["state"].values(),
            ):
                for (trained_param, trained_tensor), (
                    expected_param,
                    expected_tensor,
                ) in zip(trained.items(), expected.items()):
                    assert trained_param == expected_param
                    assert torch.allclose(
                        trained_tensor,
                        expected_tensor,
                        atol=1e-03,
                    )
            assert (
                trained_model[optimizer_part]["param_groups"]
                == expected_model[optimizer_part]["param_groups"]
            )

        # Check the encoder and decoder weights
        for model_part in ["encoder_state_dict", "decoder_state_dict"]:
            for (trained_name, trained_layer), (expected_name, expected_layer) in zip(
                trained_model[model_part].items(), expected_model[model_part].items()
            ):
                assert trained_name == expected_name
                assert torch.allclose(
                    trained_layer,
                    expected_layer,
                    atol=1e-03,
                )

        # Check the other information
        for elt in [
            "epoch",
            "step",
            "scaler_state_dict",
            "best",
            "charset",
            "curriculum_config",
        ]:
            assert trained_model[elt] == expected_model[elt]

    # Check that the inference parameters file is correct
    res = yaml.safe_load(
        (
            tmp_path
            / training_config["training"]["output_folder"]
            / "results"
            / "inference_parameters.yml"
        ).read_text()
    )
    assert res == params_res



# Copyright Teklia (contact@teklia.com) & Denis Coquenet
# This code is licensed under CeCILL-C

# -*- coding: utf-8 -*-
import re
from collections import defaultdict
from operator import attrgetter
from pathlib import Path
from typing import Dict, List, NamedTuple

import editdistance
import numpy as np

from dan.utils import parse_tokens

# Remove punctuation
REGEX_PUNCTUATION = re.compile(r"([\[\]{}/\\()\"'&+*=<>?.;:,!\-—_€#%°])")
# Remove consecutive linebreaks
REGEX_CONSECUTIVE_LINEBREAKS = re.compile(r"\n+")
# Remove consecutive spaces
REGEX_CONSECUTIVE_SPACES = re.compile(r" +")
# Keep only one space character
REGEX_ONLY_ONE_SPACE = re.compile(r"\s+")

# Mapping between computation tasks (CER, WER, NER) and their metric keyword
METRICS_KEYWORD = {"cer": "chars", "wer": "words", "ner": "tokens"}


class Inference(NamedTuple):
    """
    Store a prediction with its ground truth to avoid
    inferring again when we need to compute new metrics
    """

    image: str
    ground_truth: str
    prediction: str
    lm_prediction: str
    wer: float


class MetricManager:
    def __init__(self, metric_names: List[str], dataset_name: str, tokens: Path | None):
        self.dataset_name: str = dataset_name
        self.remove_tokens: str = None

        if tokens:
            tokens = parse_tokens(tokens)
            layout_tokens = "".join(
                list(map(attrgetter("start"), tokens.values()))
                + list(map(attrgetter("end"), tokens.values()))
            )
            self.remove_tokens: re.Pattern = re.compile(r"([" + layout_tokens + "])")
            self.keep_tokens: re.Pattern = re.compile(r"([^" + layout_tokens + "])")

        self.metric_names: List[str] = metric_names
        self.epoch_metrics = defaultdict(list)

    def format_string_for_cer(self, text: str, remove_token: bool = False):
        """
        Format string for CER computation: remove layout tokens and extra spaces
        """
        if remove_token and self.remove_tokens is not None:
            text = self.remove_tokens.sub("", text)

        text = REGEX_CONSECUTIVE_LINEBREAKS.sub("\n", text)
        return REGEX_CONSECUTIVE_SPACES.sub(" ", text).strip()

    def format_string_for_wer(
        self, text: str, remove_punct: bool = False, remove_token: bool = False
    ):
        """
        Format string for WER computation: remove layout tokens, treat punctuation as word, replace line break by space
        """
        if remove_punct:
            text = REGEX_PUNCTUATION.sub("", text)
        if remove_token and self.remove_tokens is not None:
            text = self.remove_tokens.sub("", text)
        return REGEX_ONLY_ONE_SPACE.sub(" ", text).strip().split(" ")

    def format_string_for_ner(self, text: str):
        """
        Format string for NER computation: only keep layout tokens
        """
        return self.keep_tokens.sub("", text)

    def _format_string(self, task: str, *args, **kwargs):
        """
        Call the proper `format_string_for_*` method for the given task
        """
        match task:
            case "cer":
                return self.format_string_for_cer(*args, **kwargs)
            case "wer":
                return self.format_string_for_wer(*args, **kwargs)
            case "ner":
                return self.format_string_for_ner(*args, **kwargs)

    def update_metrics(self, batch_metrics):
        """
        Add batch metrics to the metrics
        """
        for key in batch_metrics:
            self.epoch_metrics[key] += batch_metrics[key]

    def get_display_values(self, output: bool = False):
        """
        Format metrics values for shell display purposes
        """
        metric_names = self.metric_names.copy()
        if output:
            metric_names.append("nb_samples")
        display_values = dict()
        for metric_name in metric_names:
            match metric_name:
                case "time" | "nb_samples":
                    if not output:
                        continue
                    value = int(np.sum(self.epoch_metrics[metric_name]))
                    if metric_name == "time":
                        sample_time = value / np.sum(self.epoch_metrics["nb_samples"])
                        display_values["sample_time"] = float(round(sample_time, 4))
                    display_values[metric_name] = value
                    continue
                case (
                    "cer"
                    | "cer_no_token"
                    | "wer"
                    | "wer_no_punct"
                    | "wer_no_token"
                    | "ner"
                ):
                    keyword = METRICS_KEYWORD[metric_name[:3]]
                    suffix = metric_name[3:]
                    num_name, denom_name = (
                        "edit_" + keyword + suffix,
                        "nb_" + keyword + suffix,
                    )
                case "loss" | "loss_ce":
                    display_values[metric_name] = round(
                        float(
                            np.average(
                                self.epoch_metrics[metric_name],
                                weights=np.array(self.epoch_metrics["nb_samples"]),
                            ),
                        ),
                        4,
                    )
                    continue
                case _:
                    continue

            value = float(
                np.sum(self.epoch_metrics[num_name])
                / np.sum(self.epoch_metrics[denom_name])
            )
            if output:
                display_values[denom_name] = int(np.sum(self.epoch_metrics[denom_name]))
            display_values[metric_name] = round(value, 4)
        return display_values

    def compute_metrics(
        self, values: Dict[str, int | float], metric_names: List[str]
    ) -> Dict[str, List[int | float]]:
        metrics = {
            "nb_samples": [
                values["nb_samples"],
            ],
        }
        if "time" in values:
            metrics["time"] = [values["time"]]

        gt, prediction = values["str_y"], values["str_x"]
        for metric_name in metric_names:
            match metric_name:
                case (
                    "cer"
                    | "cer_no_token"
                    | "wer"
                    | "wer_no_punct"
                    | "wer_no_token"
                    | "ner"
                ):
                    task = metric_name[:3]
                    keyword = METRICS_KEYWORD[task]
                    suffix = metric_name[3:]

                    # Add extra parameters for the format functions
                    extras = []
                    if suffix == "_no_punct":
                        extras.append([{"remove_punct": True}])
                    elif suffix == "_no_token":
                        extras.append([{"remove_token": True}])

                    # Run the format function for the desired computation (CER, WER or NER)
                    split_gt = list(map(self._format_string, [task], gt, *extras))
                    split_pred = list(
                        map(self._format_string, [task], prediction, *extras)
                    )

                    # Compute and store edit distance/length for the desired level
                    # (chars, words or tokens) as metrics
                    metrics["edit_" + keyword + suffix] = list(
                        map(editdistance.eval, split_gt, split_pred)
                    )
                    metrics["nb_" + keyword + suffix] = list(map(len, split_gt))
                    metrics[keyword + "_error_rate" + suffix] = [
                        round(float(edit_dist / gt_len), 4)
                        for edit_dist, gt_len in zip(
                            metrics["edit_" + keyword + suffix],
                            metrics["nb_" + keyword + suffix],
                        )
                    ]
                case "loss" | "loss_ce":
                    metrics[metric_name] = [
                        values[metric_name],
                    ]
        return metrics

    def get(self, name: str):
        return self.epoch_metrics[name]
